#ifndef DUNE_GRIDGLUE_ADAPTER_GRIDGLUE_HH
#define DUNE_GRIDGLUE_ADAPTER_GRIDGLUE_HH

#warning This header is deprecated, use <dune/grid-glue/gridglue.hh>

#include <dune/grid-glue/gridglue.hh>

using Dune::GridGlue::GridGlue;

#endif
